/*
fmt - text formatting utility
Taylor Gurney, 2022-2024.
Public domain.

I use this program to format emails before I send them. Also handy for
HTML and really any plain-text writing.

This program accepts only one optional file argument, and no options.
It does the following transformations on the input text:
- Hard-wraps lines at 72 characters.
- Ends sentences with a single space.
- Replaces a string of whitespace characters in a line with one single
  space.
- Strips all leading and trailing spaces, removing any indentation.
*/

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "util.h"

#define HUGELEN 4096
#define WRAP 72

static char* argv0;
static char buf[HUGELEN];

/* This flag signals that the last line needs a newline added to it. */
static int neednl = 0;

static int space(char c) {
    return (c == ' ' || c == '\t' || c == '\n' || c == '\r'
            || c == '\v' || c == '\f');
}

/* remove trailing newline and return it if it is there
   otherwise return 0 */
static char chomp(char* s) {
    while (*s && *s != '\n') {
        s++;
    }
    if (*s) {
        *s = 0;
        return '\n';
    }
    return 0;
}

static void die(const char* s) {
    int e = errno;
    char* es;
    fwrite(argv0, strlen(argv0), 1, stderr);
    fwrite(": ", 2, 1, stderr);
    fwrite(s, strlen(s), 1, stderr);
    if (e) {
        es = strerror(e);
        fwrite(": ", 2, 1, stderr);
        fwrite(es, strlen(es), 1, stderr);
    }
    fwrite("\n", 1, 1, stderr);
    fflush(stderr);
    exit(1);
}

static char* token(char* s, size_t* count) {
    size_t c = 0;
    char* start = s;
    for (; !space(*s) && *s; s++, c++);
    *count = c;
    return start;
}

static void putnl(void) {
    fwrite("\n", 1, 1, stdout);
    neednl = 0;
}

static void cat(void) {
    do {
        fwrite(buf, 1, min(HUGELEN, strlen(buf)), stdout);
    } while (fgets(buf, HUGELEN, stdin));
}

static void usage(void) {
    fwrite("usage: ", 7, 1, stderr);
    fwrite(argv0, strlen(argv0), 1, stderr);
    fwrite(" [file]\n", 8, 1, stderr);
    exit(1);
}

int main(int argc, char* argv[]) {
    char* ptr = NULL;
    char* word = NULL;
    size_t wordlen = 0;
    size_t linelen = 0;
    int wordsin = 0;
    FILE* infile = stdin;
    char* inpath = "-";

    argv0 = argv[0];
    shift();
    dashdash();

    if (argc > 1) {
        usage();
    } else if (argc) {
        inpath = *argv;
        if (inpath[0] != '-' || inpath[1] != '\0') {
            infile = fopen(inpath, "r");
            if (infile == NULL) {
                die(inpath);
            }
        }
    }
    
    memset(buf, 0, HUGELEN);

    while (!feof(infile)) {
        if (fgets(buf, HUGELEN, infile) == NULL) {
            break;
        }
        /* if ultralong line then stop trying. */
        if (!chomp(buf)) {
            cat();
            return 0;
        }
        ptr = buf;
        wordsin = 0;
        for (;;) {
            for (; space(*ptr); ptr++);
            if (!*ptr) {
                if (!wordsin) {
                    if (linelen) {
                        putnl();
                        linelen = 0;
                    }
                    putnl();
                }
                break;
            }
            word = token(ptr, &wordlen);
            wordsin++;
            if (linelen) {
                if (linelen + wordlen >= WRAP) {
                    putnl();
                    linelen = 0;
                } else {
                    fwrite(" " , 1, 1, stdout);
                    neednl = 1;
                }
            }
            if (fwrite(word, wordlen, 1, stdout) != 1) {
                die("write error");
                return 1;
            }
                
            neednl = 1;
            if (linelen) {
                linelen++;
            }
            linelen += wordlen;
            ptr += wordlen;
        }
    }
    if (ferror(infile)) {
        die("read error");
        return 1;
    }
    if (neednl) {
        putnl();
    }
    return 0;
}
