/*
segoff - 21-bit x86 memory address converter
Taylor Gurney, 2022-2024.
Public domain.

This program converts x86 memory addresses between flat and
segment:offset forms.

Conversion of segmented to flat works how you would expect.  Going the
other way is more interesting, since one flat address can be equivalent
to up to 4096 different segment:offset pairs. You have the option to see
one of these pairs, or all of them.

By default, this tool operates as if on an 8086/8088, or on a later CPU
with the A20 line disabled. That is, the highest accessible address is
0xffff0. Any segment:offset pair that would go higher than that will
wrap around. To enable the A20 line, pass the -h (high) flag. This
allows you to get at the high memory area from 0x100000 to 0x10ffef. In
any case, ffff:ffff is as high as you can go.
*/

#include <limits.h>

/* for puts and fputs */
#include <stdio.h>

#include "numtab.h"
#include "util.h"
#include "u8.h"
#include "u32.h"

static int a20 = 0;
static int all = 0;

static int usage(void) {
    puts("usage: segoff [-ah] addr");
    puts("  -a  output all equivalent addresses");
    puts("  -h  enable A20 line");
    return 1;
}

static int hextou32(const char* s, u32* out) {
    const char* ptr = s;
    u32 tmp = 0;
    u32 oldtmp = 0;
    u8 digit = 0;
    if (!s) {
        return -1;
    }
    /* skip 0x */
    if (*ptr == '0' && (*(ptr+1) == 'x' || *(ptr+1) == 'X')) {
        ptr += 2;
    }
    /* ensure there is something left */
    if (!*ptr) {
        return -1;
    }
    /* skip leading 0 */
    while (*ptr == '0') {
        ptr++;
    }
    for (; *ptr; ptr++) {
        oldtmp = tmp;
        tmp <<= 4;
        digit = numtab[(u8)(*ptr)];
        if (digit > 16) {
            if ((*ptr == 'h' || *ptr == 'H') && *(ptr+1) == 0) {
                /* trailing h allowed */
                goto end;
            }
            return -1;
        }
        tmp += digit;
        if (oldtmp > tmp) {
            /* too big */
            return -1;
        }
    }
end:
    *out = tmp;
    return 0;
}

static char* u32tohex(u32 n) {
    static char buf[8];
    int i = 0;
    u32 low = 0;
    for (i = 7; i >= 0; i--) {
        low = n & 0xf;
        if (low < 10) {
            buf[i] = (char)('0' + low);
        } else {
            buf[i] = (char)('a' + low - 10);
        }
        n >>= 4;
    }
    return buf;
}

static void segoff_to_flat(u32 seg, u32 off) {
    u32 addr = seg * 0x10 + off;
    int ind = 3;
    if (!a20) {
        addr &= 0xfffff;
    } else if (addr & 0x100000) {
        ind--;
    }
    fputs(u32tohex(addr) + ind, stdout);
    puts("");
}

static void print_segoff(u32 seg, u32 off) {
    fputs(u32tohex(seg) + 4, stdout);
    fputs(":", stdout);
    fputs(u32tohex(off) + 4, stdout);
    puts("");
}

static void flat_to_segoff(u32 addr) {
    u32 seg = 0xffff;
    u32 off = addr - 0xffff0;
    if (addr < 0x100000) {
        seg = (addr & 0xf0000) >> 4;
        off = (addr & 0xffff);
    }
    print_segoff(seg, off);
}

static void flat_to_segoff_all(u32 addr) {
    u32 seg = (addr & 0x1ffff0) >> 4;
    u32 off = addr & 0xf;
    for (; seg > 0xffff; seg--, off += 16);
    for (; seg*16 + off == addr && off <= 0xffff; seg--, off += 16) {
        if (seg <= 0xffff) {
            print_segoff(seg, off);
        }
        if (!seg) {
            break;
        }
    }
}

int main(int argc, char* argv[]) {
    char* split;
    u32 addr;
    u32 seg;
    u32 off;
    argc--; argv++;
    for (; argc && argv[0][0] == '-' && argv[0][1]; argc--, argv++) {
        if (argv[0][1] == '-' && !argv[0][2]) {
            argc--;
            argv++;
            break;
        }
        for (argv[0]++; argv[0][0]; argv[0]++) {
            switch (argv[0][0]) {
            case 'a':
                all = 1;
                break;
            case 'h':
                a20 = 1;
                break;
            default:
                return usage();
            }
        }
    }
    if (argc != 1) {
        return usage();
    }
    for (split = *argv; *split && *split != ':'; split++);
    if (split && *split == ':') {
        *split = '\0';
        if (hextou32(*argv, &seg) < 0 || seg > 0xffff) {
            return usage();
        }
        if (hextou32(split+1, &off) < 0 || off > 0xffff) {
            return usage();
        }
        segoff_to_flat(seg, off);
    } else {
        if (hextou32(*argv, &addr) < 0) {
            return usage();
        }
        if (addr > 0xffff + 0xffff0) {
            puts("out of range");
            return 1;
        }
        if (!a20 && addr > 0xfffff) {
            if (addr <= 0xffff + 0xffff0) {
                puts("out of range (need a20 line)");
            } else {
                puts("out of range");
            }
            return 1;
        }
        if (all) {
            flat_to_segoff_all(addr);
        } else {
            flat_to_segoff(addr);
        }
    }
    return 0;
}

