/* 
 * Convert UTF-8 text to code page 437 (IBM PC)
 * Taylor Gurney, 2024.
 * Public domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"

typedef unsigned char u8;

#define LEN(x) (sizeof(x) / sizeof(*(x)))

struct row {
    unsigned char cp437;
    unsigned char utf8[3];
    unsigned char ascii;
};

static struct row table[] = {
    {1, {0xe2, 0x98, 0xba}, '@'}, /* "☺" */
    {2, {0xe2, 0x98, 0xbb}, '@'}, /* "☻" */
    {3, {0xe2, 0x99, 0xa5}, 'H'}, /* "♥" */
    {4, {0xe2, 0x99, 0xa6}, 'D'}, /* "♦" */
    {5, {0xe2, 0x99, 0xa3}, 'C'}, /* "♣" */
    {6, {0xe2, 0x99, 0xa0}, 'S'}, /* "♠" */
    {7, {0xe2, 0x80, 0xa2}, 'o'}, /* "•" */
    {8, {0xe2, 0x97, 0x98}, 'o'}, /* "◘" */
    {9, {0xe2, 0x97, 0x8b}, 'O'}, /* "○" */
    {10, {0xe2, 0x97, 0x99}, 'O'}, /* "◙" */
    {11, {0xe2, 0x99, 0x82}, 'M'}, /* "♂" */
    {12, {0xe2, 0x99, 0x80}, 'F'}, /* "♀" */
    {13, {0xe2, 0x99, 0xaa}, '?'}, /* "♪" */
    {14, {0xe2, 0x99, 0xab}, '?'}, /* "♫" */
    {15, {0xe2, 0x98, 0xbc}, '*'}, /* "☼" */
    {16, {0xe2, 0x96, 0xba}, '>'}, /* "►" */
    {17, {0xe2, 0x97, 0x84}, '<'}, /* "◄" */
    {18, {0xe2, 0x86, 0x95}, '|'}, /* "↕" */
    {19, {0xe2, 0x80, 0xbc}, '!'}, /* "‼" */
    {20, {0xc2, 0xb6, 0x00}, ';'}, /* "¶" */
    {21, {0xc2, 0xa7, 0x00}, 'S'}, /* "§" */
    {22, {0xe2, 0x96, 0xac}, '='}, /* "▬" */
    {23, {0xe2, 0x86, 0xa8}, '|'}, /* "↨" */
    {24, {0xe2, 0x86, 0x91}, '^'}, /* "↑" */
    {25, {0xe2, 0x86, 0x93}, 'v'}, /* "↓" */
    {26, {0xe2, 0x86, 0x92}, '>'}, /* "→" */
    {27, {0xe2, 0x86, 0x90}, '<'}, /* "←" */
    {28, {0xe2, 0x88, 0x9f}, 'L'}, /* "∟" */
    {29, {0xe2, 0x86, 0x94}, '-'}, /* "↔" */
    {30, {0xe2, 0x96, 0xb2}, '^'}, /* "▲" */
    {31, {0xe2, 0x96, 0xbc}, 'v'}, /* "▼" */
    /* ASCII printables */
    {127, {0xe2, 0x8c, 0x82}, '^'}, /* "⌂" */
    {128, {0xc3, 0x87, 0x00}, 'C'}, /* "Ç" */
    {129, {0xc3, 0xbc, 0x00}, 'u'}, /* "ü" */
    {130, {0xc3, 0xa9, 0x00}, 'e'}, /* "é" */
    {131, {0xc3, 0xa2, 0x00}, 'a'}, /* "â" */
    {132, {0xc3, 0xa4, 0x00}, 'a'}, /* "ä" */
    {133, {0xc3, 0xa0, 0x00}, 'a'}, /* "à" */
    {134, {0xc3, 0xa5, 0x00}, 'a'}, /* "å" */
    {135, {0xc3, 0xa7, 0x00}, 'c'}, /* "ç" */
    {136, {0xc3, 0xaa, 0x00}, 'e'}, /* "ê" */
    {137, {0xc3, 0xab, 0x00}, 'e'}, /* "ë" */
    {138, {0xc3, 0xa8, 0x00}, 'e'}, /* "è" */
    {139, {0xc3, 0xaf, 0x00}, 'i'}, /* "ï" */
    {140, {0xc3, 0xae, 0x00}, 'i'}, /* "î" */
    {141, {0xc3, 0xac, 0x00}, 'i'}, /* "ì" */
    {142, {0xc3, 0x84, 0x00}, 'A'}, /* "Ä" */
    {143, {0xc3, 0x85, 0x00}, 'A'}, /* "Å" */
    {144, {0xc3, 0x89, 0x00}, 'E'}, /* "É" */
    {145, {0xc3, 0xa6, 0x00}, 'e'}, /* "æ" */
    {146, {0xc3, 0x86, 0x00}, 'e'}, /* "Æ" */
    {147, {0xc3, 0xb4, 0x00}, 'o'}, /* "ô" */
    {148, {0xc3, 0xb6, 0x00}, 'o'}, /* "ö" */
    {149, {0xc3, 0xb2, 0x00}, 'o'}, /* "ò" */
    {150, {0xc3, 0xbb, 0x00}, 'u'}, /* "û" */
    {151, {0xc3, 0xb9, 0x00}, 'u'}, /* "ù" */
    {152, {0xc3, 0xbf, 0x00}, 'y'}, /* "ÿ" */
    {153, {0xc3, 0x96, 0x00}, 'O'}, /* "Ö" */
    {154, {0xc3, 0x9c, 0x00}, 'U'}, /* "Ü" */
    {155, {0xc2, 0xa2, 0x00}, 'c'}, /* "¢" */
    {156, {0xc2, 0xa3, 0x00}, 'P'}, /* "£" */
    {157, {0xc2, 0xa5, 0x00}, 'Y'}, /* "¥" */
    {158, {0xe2, 0x82, 0xa7}, 'P'}, /* "₧" */
    {159, {0xc6, 0x92, 0x00}, 'f'}, /* "ƒ" */
    {160, {0xc3, 0xa1, 0x00}, 'a'}, /* "á" */
    {161, {0xc3, 0xad, 0x00}, 'i'}, /* "í" */
    {162, {0xc3, 0xb3, 0x00}, 'o'}, /* "ó" */
    {163, {0xc3, 0xba, 0x00}, 'u'}, /* "ú" */
    {164, {0xc3, 0xb1, 0x00}, 'n'}, /* "ñ" */
    {165, {0xc3, 0x91, 0x00}, 'N'}, /* "Ñ" */
    {166, {0xc2, 0xaa, 0x00}, 'a'}, /* "ª" */
    {167, {0xc2, 0xba, 0x00}, 'o'}, /* "º" */
    {168, {0xc2, 0xbf, 0x00}, '?'}, /* "¿" */
    {169, {0xe2, 0x8c, 0x90}, '<'}, /* "⌐" */
    {170, {0xc2, 0xac, 0x00}, '>'}, /* "¬" */
    {171, {0xc2, 0xbd, 0x00}, '2'}, /* "½" */
    {172, {0xc2, 0xbc, 0x00}, '4'}, /* "¼" */
    {173, {0xc2, 0xa1, 0x00}, 'i'}, /* "¡" */
    {174, {0xc2, 0xab, 0x00}, '<'}, /* "«" */
    {175, {0xc2, 0xbb, 0x00}, '>'}, /* "»" */
    {176, {0xe2, 0x96, 0x91}, '%'}, /* "░" */
    {177, {0xe2, 0x96, 0x92}, '#'}, /* "▒" */
    {178, {0xe2, 0x96, 0x93}, '@'}, /* "▓" */
    {179, {0xe2, 0x94, 0x82}, '|'}, /* "│" */
    {180, {0xe2, 0x94, 0xa4}, '+'}, /* "┤" */
    {181, {0xe2, 0x95, 0xa1}, '+'}, /* "╡" */
    {182, {0xe2, 0x95, 0xa2}, '+'}, /* "╢" */
    {183, {0xe2, 0x95, 0x96}, '+'}, /* "╖" */
    {184, {0xe2, 0x95, 0x95}, '+'}, /* "╕" */
    {185, {0xe2, 0x95, 0xa3}, '+'}, /* "╣" */
    {186, {0xe2, 0x95, 0x91}, '|'}, /* "║" */
    {187, {0xe2, 0x95, 0x97}, '+'}, /* "╗" */
    {188, {0xe2, 0x95, 0x9d}, '+'}, /* "╝" */
    {189, {0xe2, 0x95, 0x9c}, '+'}, /* "╜" */
    {190, {0xe2, 0x95, 0x9b}, '+'}, /* "╛" */
    {191, {0xe2, 0x94, 0x90}, '+'}, /* "┐" */
    {192, {0xe2, 0x94, 0x94}, '+'}, /* "└" */
    {193, {0xe2, 0x94, 0xb4}, '+'}, /* "┴" */
    {194, {0xe2, 0x94, 0xac}, '+'}, /* "┬" */
    {195, {0xe2, 0x94, 0x9c}, '+'}, /* "├" */
    {196, {0xe2, 0x94, 0x80}, '-'}, /* "─" */
    {197, {0xe2, 0x94, 0xbc}, '+'}, /* "┼" */
    {198, {0xe2, 0x95, 0x9e}, '+'}, /* "╞" */
    {199, {0xe2, 0x95, 0x9f}, '+'}, /* "╟" */
    {200, {0xe2, 0x95, 0x9a}, '+'}, /* "╚" */
    {201, {0xe2, 0x95, 0x94}, '+'}, /* "╔" */
    {202, {0xe2, 0x95, 0xa9}, '+'}, /* "╩" */
    {203, {0xe2, 0x95, 0xa6}, '+'}, /* "╦" */
    {204, {0xe2, 0x95, 0xa0}, '|'}, /* "╠" */
    {205, {0xe2, 0x95, 0x90}, '='}, /* "═" */
    {206, {0xe2, 0x95, 0xac}, '+'}, /* "╬" */
    {207, {0xe2, 0x95, 0xa7}, '+'}, /* "╧" */
    {208, {0xe2, 0x95, 0xa8}, '+'}, /* "╨" */
    {209, {0xe2, 0x95, 0xa4}, '+'}, /* "╤" */
    {210, {0xe2, 0x95, 0xa5}, '+'}, /* "╥" */
    {211, {0xe2, 0x95, 0x99}, '+'}, /* "╙" */
    {212, {0xe2, 0x95, 0x98}, '+'}, /* "╘" */
    {213, {0xe2, 0x95, 0x92}, '+'}, /* "╒" */
    {214, {0xe2, 0x95, 0x93}, '+'}, /* "╓" */
    {215, {0xe2, 0x95, 0xab}, '+'}, /* "╫" */
    {216, {0xe2, 0x95, 0xaa}, '+'}, /* "╪" */
    {217, {0xe2, 0x94, 0x98}, '+'}, /* "┘" */
    {218, {0xe2, 0x94, 0x8c}, '+'}, /* "┌" */
    {219, {0xe2, 0x96, 0x88}, '#'}, /* "█" */
    {220, {0xe2, 0x96, 0x84}, '-'}, /* "▄" */
    {221, {0xe2, 0x96, 0x8c}, '|'}, /* "▌" */
    {222, {0xe2, 0x96, 0x90}, '|'}, /* "▐" */
    {223, {0xe2, 0x96, 0x80}, '-'}, /* "▀" */
    {224, {0xce, 0xb1, 0x00}, 'a'}, /* "α" */
    {225, {0xc3, 0x9f, 0x00}, 's'}, /* "ß" */
    {226, {0xce, 0x93, 0x00}, 'G'}, /* "Γ" */
    {227, {0xcf, 0x80, 0x00}, 'p'}, /* "π" */
    {228, {0xce, 0xa3, 0x00}, 'S'}, /* "Σ" */
    {229, {0xcf, 0x83, 0x00}, 's'}, /* "σ" */
    {230, {0xc2, 0xb5, 0x00}, 'u'}, /* "µ" */
    {231, {0xcf, 0x84, 0x00}, 't'}, /* "τ" */
    {232, {0xce, 0xa6, 0x00}, 'f'}, /* "Φ" */
    {233, {0xce, 0x98, 0x00}, 't'}, /* "Θ" */
    {234, {0xce, 0xa9, 0x00}, 'W'}, /* "Ω" */
    {235, {0xce, 0xb4, 0x00}, 'd'}, /* "δ" */
    {236, {0xe2, 0x88, 0x9e}, '%'}, /* "∞" */
    {237, {0xcf, 0x86, 0x00}, 'p'}, /* "φ" */
    {238, {0xce, 0xb5, 0x00}, 'e'}, /* "ε" */
    {239, {0xe2, 0x88, 0xa9}, 'A'}, /* "∩" */
    {240, {0xe2, 0x89, 0xa1}, '='}, /* "≡" */
    {241, {0xc2, 0xb1, 0x00}, '+'}, /* "±" */
    {242, {0xe2, 0x89, 0xa5}, '>'}, /* "≥" */
    {243, {0xe2, 0x89, 0xa4}, '<'}, /* "≤" */
    {244, {0xe2, 0x8c, 0xa0}, '|'}, /* "⌠" */
    {245, {0xe2, 0x8c, 0xa1}, '|'}, /* "⌡" */
    {246, {0xc3, 0xb7, 0x00}, '+'}, /* "÷" */
    {247, {0xe2, 0x89, 0x88}, '='}, /* "≈" */
    {248, {0xc2, 0xb0, 0x00}, 'o'}, /* "°" */
    {249, {0xe2, 0x88, 0x99}, 'o'}, /* "∙" */
    {250, {0xc2, 0xb7, 0x00}, '*'}, /* "·" */
    {251, {0xe2, 0x88, 0x9a}, '/'}, /* "√" */
    {252, {0xe2, 0x81, 0xbf}, 'n'}, /* "ⁿ" */
    {253, {0xc2, 0xb2, 0x00}, '2'}, /* "²" */
    {254, {0xe2, 0x96, 0xa0}, '#'}, /* "■" */
};


static unsigned char to_cp437(unsigned char* seq) {
    size_t j;
    for (j = 0; j < LEN(table); j++) {
        if (memcmp(table[j].utf8, seq, 3) == 0) {
            return table[j].cp437;
        }
    }
    /* fail */
    return 0;
}

#define BUFLEN 5
int main(int argc, char* argv[]) {
    char* path = NULL;
    FILE* f = NULL;
    unsigned char buf[BUFLEN];
    size_t n = 0;
    size_t i = 0;
    unsigned char out;
    if (argc != 2) {
        fputs("usage: ", stderr);
        fputs(argv[0], stderr);
        fputs(" <file>\n", stderr);
        exit(1);
    }
    path = argv[1];
    f = fopen(path, "rb");
    if (!f) {
        fputs(argv[0], stderr);
        fputs(": ", stderr);
        perror(path);
        exit(1);
    }
    do {
        n = fread(buf, 1, 1, f);
        if (n != 1) break;
        memset(buf+1, 0, BUFLEN-1);
        if (*buf <= 0x7f) {
            fwrite(&buf, 1, 1, stdout);
            continue;
        }
        /* read next */
        for (i = 1; i < 3; i++) {
            n = fread(buf+i, 1, 1, f);
            memset(buf+i+1, 0, BUFLEN-i-1);
            if (!n) goto fail;
            out = to_cp437(buf);
            if (out) {
                fwrite(&out, 1, 1, stdout);
                goto next;
            }
        }
fail:
        fputs(argv[0], stderr);
        fputs(": invalid char sequence in file\n", stderr);
        exit(1);
next: ;
    } while (n);
    if (ferror(f)) {
        perror(path);
        exit(1);
    }
    return 0;
}

