/*
xd - a small and fast hex dump tool
Written by Taylor Gurney, 2022-2024.
Public domain.
*/

/*

xd is a simple hex dump program. Its output and functionality are similar to
that of the common "xxd" program, including a reverse mode that can be used for
binary file patching. It can read xxd's default-format hex dumps as well.

xd is written in ISO standard C89. I have tested it on FreeDOS, Linux, and
MacOS.

I'm aware that there have been other hex dump tools by the same name released
over the years. This "xd" is unrelated to any others by that name.

=== SYNOPSIS ===

xd [-r] [-j n] [-l n] [-o n] [infile [outfile]]

=== DESCRIPTION ===

xd creates a hex dump of a specified file, or, in its alternate mode, reverses
a hex dump back into raw binary output.

By default, if no file is specified, xd reads data from standard input, and
writes a hex dump of that data to standard output. Each output line begins with
a 32-bit file offset, then a data portion with up to 16 bytes written as pairs
of hex digits, and finally a printable ASCII dump.

When the -r flag is given, xd operates in reverse mode. The input (whether from
standard input or from a specified file name) must be a hex dump produced by xd
or a compatible tool. The output will be the raw binary that corresponds to the
hex dump.  xd honors the offset at the beginning of each line, and will seek to
that offset if it is not the current offset in the output file. Both forward
and backward seeks are allowed if the output file is seekable. Also, when a
named output file (not standard input) is specified in this mode, xd will not
truncate it. This can be used to patch binary files.

=== OPTIONS ===

-j n  Skip the first n input bytes.

-l n  Stop after writing n output bytes.

-o n  Add n to all offsets. The number may be negative.

-r    Reverse mode.

=== EXIT STATUS ===

 0  Successful completion.

>0  An error occurred, or, some malformed input was found while running in
reverse mode.

=== AUTHOR ===

Taylor Gurney

=== CAVEATS ===

Big numbers:

All numbers used in calculation are 32-bit unsigned.

Junk:

In reverse mode, any "junk" found in an input line during parsing will cause xd
to discard the remainder of the that line and move on to the next line.

Seeking:

Seeking past the end of an output file may fill the gap with null-bytes, or it
may create a "hole" if the file system supports sparse files, or it may just
seek to the end of the file. Or it may do something else entirely. What happens
depends on the operating system and file system in use.

Seeking is not supported in non-seekable output files. Examples: pipes, FIFOs,
sockets, TTYs.
*/

/* for errno */
#include <errno.h>

/* for fwrite, fflush, fopen, fclose, fputs, stdin, stdout, stderr, FILE, rewind */
#include <stdio.h>

/* for exit */
#include <stdlib.h>

/* for memset, strlen, strerror */
#include <string.h>

#include "numtab.h"
#include "strtou32.h"
#include "u32.h"
#include "u8.h"
#include "util.h"

#define HEX_BYTES_PER_ROW 16
#define OFFSET_LEN 8
#define ROW_LEN (HEX_BYTES_PER_ROW * 4 + OFFSET_LEN + 5)

static char* argv0;
static char* inpath = "-";
static char* outpath = "-";
static const char* hextab = "0123456789abcdef";

static void die(const char* s) {
    int e = errno;
    char* es;
    fwrite(argv0, strlen(argv0), 1, stderr);
    fwrite(": ", 2, 1, stderr);
    fwrite(s, strlen(s), 1, stderr);
    if (e) {
        es = strerror(e);
        fwrite(": ", 2, 1, stderr);
        fwrite(es, strlen(es), 1, stderr);
    }
    fwrite("\n", 1, 1, stderr);
    fflush(stderr);
    exit(1);
}

static void usage(void) {
    fputs("Usage: \n  ", stderr);
    fputs(argv0, stderr);
    fputs(
" [-r] [-j -n] [-l n] [-o n] [infile [outfile]]\n"
"\nOptions:\n"
"  -j n  skip first <n> input bytes\n"
"  -l n  stop after <n> bytes\n"
"  -o n  add <n> to all offsets (can be negative)\n"
"  -r    reverse mode\n",
    stderr);
    exit(1);
}

static u8 hextonum(char* digits, u32* out) {
    u8 c = 0;
    u8 i = 0;
    *out = 0;
    for (i = 0; i < OFFSET_LEN; i++) {
        c = numtab[(u8)digits[i]];
        if (c > 15) {
            break;
        }
        *out <<= 4;
        *out |= c;
    }
    return i;
}

static void numtohex(u32 num, u8* out, int digits) {
    while (digits--) {
        *(out + digits) = (u8)hextab[num & 0xf];
        num >>= 4;
    }
}

static void xd(FILE* inf, FILE* outf, u32 writelen,
        u32 skip, u32 adjust) {
    u8 buf[HEX_BYTES_PER_ROW];
    size_t bytesread = 0;
    size_t i;
    u32 count = 0;
    u8 outrow[ROW_LEN] = {0};
    const u8 off4bit = skip & 0xf;
    while (skip) {
        bytesread = fread(buf, 1, min(HEX_BYTES_PER_ROW, skip), inf);
        if (!bytesread) {
            if (ferror(inf)) {
                die(inpath);
            }
            return;
        }
        skip -= bytesread;
        count += bytesread;
    }
    bytesread = 0;
    memset(outrow, ' ', ROW_LEN);
    outrow[OFFSET_LEN] = ':';
    outrow[ROW_LEN - 1] = '\0';
    outrow[ROW_LEN - 2] = '\n';
    while (!feof(inf) && writelen) {
        bytesread = fread(buf, 1, HEX_BYTES_PER_ROW, inf);
        if (bytesread < 1 && ferror(inf)) {
            die(inpath);
        }
        if (bytesread) {
            numtohex(count + adjust, outrow, OFFSET_LEN);
        }
        bytesread = min(writelen, bytesread);
        for (i = 0; i < bytesread; i++) {
            count++;
            numtohex((u32)buf[i], outrow + OFFSET_LEN + 2 + i * 3, 2);
            outrow[ROW_LEN - HEX_BYTES_PER_ROW - 2 + i] = bintoprint(buf[i]);
            if (count % HEX_BYTES_PER_ROW == off4bit) {
                fputs((char*)outrow, outf);
            }
        }
        writelen -= bytesread;
    }
    /* last row */
    count = (count - off4bit) % HEX_BYTES_PER_ROW;
    if (count && bytesread) {
        for (i = count; i < HEX_BYTES_PER_ROW; i++) {
            memset(outrow + OFFSET_LEN + 2 + i * 3, ' ', 3);
            outrow[ROW_LEN - HEX_BYTES_PER_ROW - 2 + i] = ' ';
        }
        fputs((char*)outrow, outf);
    }
}

static int rxd(FILE* inf, FILE* outf, u32 writelen,
        u32 skip, u32 adjust) {
    char inbuf[ROW_LEN + 1];
    size_t i;
    int junk = 0;
    u32 fileoff;
    u32 count = 0;
    u8 outbuf[ROW_LEN];
    u32 outlen;
    u8 digit0;
    u8 digit1;
    u8 offlen = 0;
    inbuf[ROW_LEN] = '\0';
    while (!feof(inf) && writelen) {
        if (!fgets(inbuf, ROW_LEN, inf)) {
            if (ferror(inf)) {
                die(inpath);
            }
            return junk;
        }
        offlen = hextonum(inbuf, &fileoff);
        if (offlen < 1) {
            /* junk, but not erroneous.
               could be a comment or something. */
            continue;
        }
        fileoff += adjust;
        if (count != fileoff) {
            if (fseek(outf, fileoff, SEEK_SET) < 0) {
                die(outpath);
            }
            count = fileoff;
        }
        outlen = 0;
        for (i = offlen + 1; i <= ROW_LEN; i++) {
            if (inbuf[i] < ' ') {
                break;
            }
            if (inbuf[i] == ' ') {
                if (inbuf[i+1] == ' ') {
                    break;
                }
                continue;
            }
            /* hex to num */
            digit0 = numtab[(u8)inbuf[i]];
            digit1 = numtab[(u8)inbuf[i+1]];
            if ((digit0 | digit1) > 15) {
                junk = 1;
                break;
            }
            outbuf[outlen] = (u8)(digit0 << 4) | digit1;
            outlen++;
            count++;
            i++;
        }
        if (skip >= outlen) {
            skip -= outlen;
        } else {
            if (skip) {
                outlen -= skip;
            }
            outlen = min(writelen, outlen);
            if (outlen && fwrite(outbuf+skip, outlen, 1, outf) != 1
                    && ferror(outf)) {
                die(outpath);
            }
            writelen -= outlen;
            skip = 0;
        }
    }
    return junk;
}

static FILE* fopen_or_die(const char* path, const char* mode) {
    FILE* f = fopen(path, mode);
    if (f == NULL) {
        die(path);
    }
    return f;
}

int main(int argc, char* argv[]) {
    int rval = 0;
    FILE* inf = stdin;
    FILE* outf = stdout;
    int reverse = 0;
    u32 writelen = 0;
    u32 skip = 0;
    u32 adjust = 0;
    int nextarg = 0;
    char** argv_;
    writelen--; /* wrap to uint32 max */
    argv0 = argv[0];
    shift();
    for (; argc && argv[0][0] == '-' && argv[0][1]; shift()) {
        if (argv[0][1] == '-' && !argv[0][2]) {
            shift();
            break;
        }
        nextarg = 0;
        for (argv[0]++; !nextarg && argv[0][0]; argv[0]++) {
            argv_ = argv;
            switch (argv[0][0]) {
            case 'j':
                if (strtou32(optarg(), &skip, 0) != 0) {
                    die("argument to -j must be a positive integer");
                }
                break;
            case 'l':
                if (strtou32(optarg(), &writelen, 0) != 0) {
                    die("argument to -l must be a positive integer");
                }
                break;
            case 'o':
                if (strtou32(optarg(), &adjust, 1) != 0) {
                    die("argument to -o must be an integer");
                }
                break;
            case 'r':
                reverse = 1;
                break;
            default:
                usage();
            }
        }
    }

    if (argc > 0) {
        inpath = argv[0];
        if (!(inpath[0] == '-' && inpath[1] == '\0')) {
            inf = fopen_or_die(inpath, "rb");
        }
        if (argc > 1) {
            outpath = argv[1];
            if (!(outpath[0] == '-' && outpath[1] == '\0')) {
                if (reverse) {
                    /* don't truncate.
                       open in append mode first to ensure the file is created
                    */
                    outf = fopen_or_die(outpath, "ab");
                    if (fclose(outf)) {
                        die(outpath);
                    }
                    outf = fopen_or_die(outpath, "r+b");
                } else {
                    outf = fopen_or_die(outpath, "wb");
                }
                rewind(outf);
            }
        }
    }

    if (reverse) {
        rval = rxd(inf, outf, writelen, skip, adjust);
    } else {
        xd(inf, outf, writelen, skip, adjust);
    }
    fclose(inf);
    fclose(outf);
    return rval;
}
